#!/usr/bin/env python
#
# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cartola import config
import unittest
from firenado import data
import firenado.conf
from sqlalchemy.engine import base as base_engine
from sqlalchemy.future import engine as future_engine


class MockDataConnected(data.DataConnectedMixin):
    """ Data connected mock object. This object holds the data sources to be
    used in the test cases.
    """
    def __init__(self):
        pass


class SqlalchemyConnectorTestCase(unittest.TestCase):
    """ Case that tests an Firenado application after being loaded from its
    configuration file.
    """

    def setUp(self):
        """ Setting up an object that has firenado.core.service.served_by
        decorators on some methods.
        """
        self.data_connected_instance = MockDataConnected()

    def test_engine_not_future(self):
        """ Test if the engine created will be the base one if future is set
        as False or not set in the data source configuration. """
        data_source_name = "not_future_data_source"
        data_source_conf = {
            'connector': "sqlalchemy",
            'url': "mysql+pymysql://root@localhost:3306/test",
            'future': False
        }
        data_source = data.config_to_data_source(data_source_name,
                                                 data_source_conf,
                                                 self.data_connected_instance)
        self.assertTrue(isinstance(data_source.engine, base_engine.Engine))
        data_source_conf = {
            'connector': "sqlalchemy",
            'url': "mysql+pymysql://root@localhost:3306/test"
        }
        data_source = data.config_to_data_source(data_source_name,
                                                 data_source_conf,
                                                 self.data_connected_instance)
        self.assertTrue(isinstance(data_source.engine, base_engine.Engine))

    def test_engine_future(self):
        """ Test if the engine created will be the future one if future is set
        as True in the data source configuration. """
        data_source_name = "future_data_source"
        data_source_conf = {
            'connector': "sqlalchemy",
            'url': "mysql+pymysql://root@localhost:3306/test",
            'future': True
        }
        data_source = data.config_to_data_source(data_source_name,
                                                 data_source_conf,
                                                 self.data_connected_instance)
        self.assertTrue(isinstance(data_source.engine, future_engine.Engine))

    def test_parametrized_instead_url(self):
        """ Test a parametrized data source configuration will generate a
        valid sqlalchemy connection url."""
        connector_conf = firenado.conf.data['connectors']["sqlalchemy"]
        handler_class = config.get_from_module(connector_conf['module'],
                                               connector_conf['class'])
        data_source_instance = handler_class(self.data_connected_instance)

        # At this point the connector parameter doesn't matter because we
        # already have the handler class instantiated correctly.
        # Keeping it at the data_source conf just for the sake of clarity.
        url_all_parameters = "mysql+pymysql://root:apass@localhost:3306/test"
        conf = data_source_instance.process_config({
            'connector': "sqlalchemy",
            'dialect': "mysql",
            'driver': "pymysql",
            'username': "root",
            'password': "apass",
            'host': "localhost",
            'port': "3306",
            'database': "test"
        })
        self.assertEqual(url_all_parameters, conf['url'])

        url_no_password = "mysql+pymysql://root@localhost:3306/test"
        conf = data_source_instance.process_config({
            'connector': "sqlalchemy",
            'dialect': "mysql",
            'driver': "pymysql",
            'user': "root",  # using user instead username
            'host': "localhost",
            'port': "3306",
            'database': "test"
        })
        self.assertEqual(url_no_password, conf['url'])

        url_no_port = "mysql+pymysql://root:apass@localhost/test"
        conf = data_source_instance.process_config({
            'connector': "sqlalchemy",
            'dialect': "mysql",
            'driver': "pymysql",
            'username': "root",
            'pass': "apass",  # using pass instead of password
            'host': "localhost",
            'db': "test"  # using db instead of database
        })
        self.assertEqual(url_no_port, conf['url'])



# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from behave.api.async_step import async_run_until_complete
from firenado.launcher import ProcessLauncher
from tornado.testing import (bind_unused_port, AsyncTestCase,
                             AsyncHTTPTestCase)
from tornado.httpclient import HTTPResponse
from tornado import gen, ioloop
from typing import Any
import warnings


def get_event_loop():
    """ Return a new event asyncio evenvent loop with trigerring a
    depreciation warning from Python 10.

    :return: A event loop
    """
    loop = asyncio.get_event_loop_policy().get_event_loop()
    return loop if loop else asyncio.new_event_loop()


class ProcessLauncherTestCase(AsyncHTTPTestCase):

    @property
    def launcher(self):
        return self.__launcher

    def get_launcher(self) -> ProcessLauncher:
        """Should be overridden by subclasses to return a
        `firenado.launcher.ProcessLauncher`.
        """
        raise NotImplementedError()

    def get_http_port(self) -> int:
        """Returns the port used by the server.

        A new port is chosen for each test.
        """
        return self.__port

    @async_run_until_complete(should_close=False, loop=get_event_loop())
    async def setUp(self) -> None:
        self.should_close_asyncio_loop = False
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sock, port = bind_unused_port()
            sock.close()
            self.__port = port
            self.__launcher = self.get_launcher()
            self.__launcher.port = self.__port
            self.__launcher.load()
            self.http_client = self.get_http_client()
            self.io_loop = ioloop.IOLoop.current()
            await self.__launcher.launch()
            await gen.sleep(1)

    def get_url(self, path: str) -> str:
        """Returns an absolute url for the given path on the test server."""
        return "%s://127.0.0.1:%s%s" % (self.get_protocol(),
                                        self.get_http_port(), path)

    async def fetch(
        self, path: str, raise_error: bool = False, **kwargs: Any
    ) -> HTTPResponse:
        """Convenience method to synchronously fetch a URL.

        The given path will be appended to the local server's host and
        port.  Any additional keyword arguments will be passed directly to
        `.AsyncHTTPClient.fetch` (and so could be used to pass
        ``method="POST"``, ``body="..."``, etc).

        If the path begins with http:// or https://, it will be treated as a
        full URL and will be fetched as-is.

        If ``raise_error`` is ``True``, a `tornado.httpclient.HTTPError` will
        be raised if the response code is not 200. This is the same behavior
        as the ``raise_error`` argument to `.AsyncHTTPClient.fetch`, but
        the default is ``False`` here (it's ``True`` in `.AsyncHTTPClient`)
        because tests often need to deal with non-200 response codes.

        .. versionchanged:: 5.0
           Added support for absolute URLs.

        .. versionchanged:: 5.1

           Added the ``raise_error`` argument.

        .. deprecated:: 5.1

           This method currently turns any exception into an
           `.HTTPResponse` with status code 599. In Tornado 6.0,
           errors other than `tornado.httpclient.HTTPError` will be
           passed through, and ``raise_error=False`` will only
           suppress errors that would be raised due to non-200
           response codes.

        """
        if path.lower().startswith(("http://", "https://")):
            url = path
        else:
            url = self.get_url(path)

        return await self.http_client.fetch(url, raise_error=raise_error,
                                            **kwargs)

    def tearDown(self) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.http_client.close()
            AsyncTestCase.tearDown(self)
            self.__launcher.shutdown()



# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

warnings.warn("The \"firenado.util.sqlalchemy_util\" module is depreciated. "
              "Please use firenado.sqlalchemy  instead. This module will be "
              "removed on Firenado 0.9.1.", DeprecationWarning,
              2)

from ..sqlalchemy import (base_to_dict, fast_count, run_script)



# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from behave import given, when, then
from behave.api.async_step import async_run_until_complete
from tests import PROJECT_ROOT, chdir_fixture_app
from firenado.launcher import ProcessLauncher
from tornado import gen
from tornado.httpclient import AsyncHTTPClient
import sys


@given("command {command} is called")
async def command_is_called(context, command):
    context.tester.assertTrue(False)
    print("Buga")



# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from tests.service_test import TestableDataConnected, ServedByInstance
from firenado.sqlalchemy import base_to_dict, with_session
from firenado.service import FirenadoService, with_service
from sqlalchemy import String
from sqlalchemy.types import DateTime
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.sql import text
import unittest


class Base(DeclarativeBase):
    pass


class TestBase(Base):

    __tablename__ = "test"

    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(String(150), nullable=False)
    first_name: Mapped[str] = mapped_column(String(150), nullable=False)
    last_name: Mapped[str] = mapped_column(String(150), nullable=False)
    password: Mapped[str] = mapped_column(String(150), nullable=False)
    email: Mapped[str] = mapped_column(String(150), nullable=False)
    created: Mapped[datetime] = mapped_column(DateTime, nullable=False,
                                              server_default=text("now()"))
    modified: Mapped[datetime] = mapped_column(DateTime, nullable=False,
                                               server_default=text("now()"))


class MockSessionedService(FirenadoService):

    @with_session
    def resolve_from_default_data_source(self, **kwargs):
        return kwargs

    @with_session(data_source="datasource1")
    def resolve_from_data_source(self, **kwargs):
        return kwargs

    @property
    def default_data_source(self):
        return "datasource2"


class BaseToDictTestCase(unittest.TestCase):

    def setUp(self):
        self.test_object = TestBase()
        self.test_object.id = 1
        self.test_object.username = "anusername"
        self.test_object.password = "apassword"
        self.test_object.first_name = "Test"
        self.test_object.last_name = "Object"
        self.test_object.email = "test@example.com"

    def test_base_to_dict(self):
        dict_from_base = base_to_dict(self.test_object)
        self.assertEqual(dict_from_base['id'], self.test_object.id)
        self.assertEqual(dict_from_base['username'], self.test_object.username)
        self.assertEqual(dict_from_base['password'], self.test_object.password)
        self.assertEqual(dict_from_base['first_name'],
                         self.test_object.first_name)
        self.assertEqual(dict_from_base['last_name'],
                         self.test_object.last_name)
        self.assertEqual(dict_from_base['email'], self.test_object.email)
        self.assertEqual(dict_from_base['created'], self.test_object.created)
        self.assertEqual(dict_from_base['modified'], self.test_object.modified)

    def test_base_to_dict_parametrized(self):
        dict_from_base = base_to_dict(self.test_object,
                                      ["id", "username", "first_name"])
        self.assertEqual(dict_from_base['id'], self.test_object.id)
        self.assertEqual(dict_from_base['username'], self.test_object.username)
        self.assertEqual(dict_from_base['first_name'],
                         self.test_object.first_name)
        self.assertTrue("password" not in dict_from_base)
        self.assertTrue("last_name" not in dict_from_base)
        self.assertTrue("email" not in dict_from_base)
        self.assertTrue("created" not in dict_from_base)
        self.assertTrue("modified" not in dict_from_base)


class SessionedTestCase(unittest.TestCase):

    mock_sessioned_service: MockSessionedService

    def setUp(self):
        """ Setting up an object that has firenado.core.service.served_by
        decorators on some methods.
        """
        self.data_connected_instance = TestableDataConnected()
        self.served_by_instance = ServedByInstance(
            self.data_connected_instance)

    @property
    def data_connected(self):
        return self.served_by_instance.data_connected

    @with_service(MockSessionedService)
    def test_sessioned_default_data_source(self):
        """ Method resolve_from_default_data_source is anoteded with sessioned
        and no parameter. The data source to be used is the one defined either
        in the property or method named default_data_source. In this case we're
        using a property.

        As no session was provided the session will be closed
        """
        resolved_kwargs = (
            self.mock_sessioned_service.resolve_from_default_data_source()
        )
        self.assertEqual("datasource2", resolved_kwargs['data_source'])
        data_source = self.data_connected.get_data_source(
            resolved_kwargs['data_source']
        )
        self.assertEqual(data_source.resolve_session(),
                         resolved_kwargs['session'].name)
        self.assertFalse(resolved_kwargs['session'].is_oppened)

    @with_service(MockSessionedService)
    def test_sessioned_default_data_source_my_session(self):
        """ Method resolve_from_default_data_source is anoteded with sessioned
        and no parameter. Instead of getting the session from the default data
        source, we're providing our own session.
        """
        data_source = self.data_connected.get_data_source("datasource1")
        resolved_kwargs = (
            self.mock_sessioned_service.resolve_from_default_data_source(
                session=data_source.session
            )
        )
        # No datasource will be added to the kwards
        self.assertIsNone(resolved_kwargs.get("data_source"))
        # Using session provided
        self.assertEqual(data_source.resolve_session(),
                         resolved_kwargs['session'].name)
        self.assertTrue(resolved_kwargs['session'].is_oppened)

    @with_service(MockSessionedService)
    def test_sessioned_from_data_source(self):
        """ Method resolve_from_data_source is anoteded with sessioned and
        datasource1 as parameter. It will be injected to the method kwargs
        session from datasource1.
        """
        resolved_kwargs = (
            self.mock_sessioned_service.resolve_from_data_source()
        )
        self.assertEqual("datasource1", resolved_kwargs['data_source'])
        data_source = self.data_connected.get_data_source(
            resolved_kwargs['data_source']
        )
        self.assertEqual(data_source.resolve_session(),
                         resolved_kwargs['session'].name)
        self.assertFalse(resolved_kwargs['session'].is_oppened)

    @with_service(MockSessionedService)
    def test_sessioned_from_data_source_provide_session(self):
        """ Method resolve_from_default_data_source is anoteded with sessioned
        and no parameter. This time we provide the session and provided to
        close the datasource after executing resolve_from_data_source.
        """
        data_source = self.data_connected.get_data_source("datasource2")
        resolved_kwargs = (
            self.mock_sessioned_service.resolve_from_data_source(
                session=data_source.session, close=True
            )
        )
        # No datasource will be added to the kwards
        self.assertIsNone(resolved_kwargs.get("data_source"))
        # Using session provided
        self.assertEqual(data_source.resolve_session(),
                         resolved_kwargs['session'].name)
        self.assertFalse(resolved_kwargs['session'].is_oppened)

    @with_service(MockSessionedService)
    def test_sessioned_with_my_data_source_closing_connection(self):
        """ Method resolve_from_data_source is anoteded with sessioned and
        datasource1 as parameter. We're overwriting the data_source parameter
        during the method call, changing to datasource2. It will be injected
        to the method kwargs a session from data_source1.
        """
        resolved_kwargs = (
            self.mock_sessioned_service.resolve_from_data_source(
                data_source="datasource2")
        )
        self.assertEqual("datasource2", resolved_kwargs['data_source'])
        data_source = self.data_connected.get_data_source(
            resolved_kwargs['data_source']
        )
        self.assertEqual(data_source.resolve_session(),
                         resolved_kwargs['session'].name)
        self.assertFalse(resolved_kwargs['session'].is_oppened)



# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cartola import sysexits
import firenado.conf
from importlib import reload
import logging
import os
import sys

logger = logging.getLogger(__name__)


class FirenadoLauncher:

    def __init__(self, **settings):
        self.app = settings.get("app", None)
        self.path = settings.get("path", None)
        if self.path is not None:
            sys.path.append(self.path)
        if self.app:
            os.environ["CURRENT_APP"] = self.app
        if os.environ.get("CURRENT_APP"):
            self.app = os.environ.get("CURRENT_APP")
        self.addresses = settings.get("addresses", None)
        self.dir = settings.get("dir", None)
        self.port = settings.get("port", None)
        self.socket = settings.get("socket", None)

        # Fixing and cleaning PYTHONPATH and sys.path
        # This is useful so we can run a process launcher with the same
        # PYTHONPATH from the parent process
        real_pythonpaths = sys.path[:]
        if "PYTHONPATH" in os.environ and os.environ['PYTHONPATH'] is not None:
            current_pythonpaths = os.environ['PYTHONPATH'].split(":")
            for path in current_pythonpaths:
                if path.strip() != "" and path.strip() not in real_pythonpaths:
                    real_pythonpaths.append(path.strip())
            for path in sys.path:
                if path.strip() != "" and path.strip() not in real_pythonpaths:
                    real_pythonpaths.append(path.strip())
        sys.path = real_pythonpaths
        os.environ['PYTHONPATH'] = ":".join(real_pythonpaths)
        if self.dir is not None:
            os.chdir(self.dir)
        if self.app is not None or self.dir is not None:
            reload(firenado.conf)

        # Set logging basic configurations
        for handler in logging.root.handlers[:]:
            # clearing loggers, solution from: https://bit.ly/2yTchyx
            logging.root.removeHandler(handler)
        logging.basicConfig(level=firenado.conf.log['level'],
                            format=firenado.conf.log['format'])

    def load(self):
        raise NotImplementedError()

    def launch(self):
        raise NotImplementedError()

    def shutdown(self):
        raise NotImplementedError()


def expect_list_mp(self, pattern_list, timeout=-1, searchwindowsize=-1,
                   async_=False, **kw):
    from pexpect import Expecter, searcher_re
    '''This takes a list of compiled regular expressions and returns the
    index into the pattern_list that matched the child output. The list may
    also contain EOF or TIMEOUT(which are not compiled regular
    expressions). This method is similar to the expect() method except that
    expect_list() does not recompile the pattern list on every call. This
    may help if you are trying to optimize for speed, otherwise just use
    the expect() method.  This is called by expect().


    Like :meth:`expect`, passing ``async_=True`` will make this return an
    asyncio coroutine.
    '''
    if timeout == -1:
        timeout = self.timeout
    if 'async' in kw:
        async_ = kw.pop('async')
    if kw:
        raise TypeError("Unknown keyword arguments: {}".format(kw))

    exp = Expecter(self, searcher_re(pattern_list), searchwindowsize)
    if async_:
        from ._pexpect_async_patch import expect_async
        return expect_async(exp, timeout)
    else:
        return exp.expect_loop(timeout)


class ProcessLauncher(FirenadoLauncher):
    try:
        import pexpect
        process: pexpect.spawn = None
    except ImportError:
        logger.debug("The pexpect module ins't isn't installed. Consider "
                     "installing it in order to launch applications using "
                     "ProcessLauncher.")

    def __init__(self, **settings):
        super().__init__(**settings)
        self.process = None
        self.process_callback = None
        self.process_callback_time = 100
        self.logfile = settings.get("logfile", None)
        self.command = None
        self.response = None

    def load(self):
        firenado_script = os.path.join(firenado.conf.ROOT, "bin",
                                       "firenado-cli.py")
        self.command = f"{sys.executable} {firenado_script} app run"
        if self.dir is not None:
            self.command = f"{self.command} --dir={self.dir}"
        if self.socket is None and self.addresses is not None:
            self.command = f"{self.command} --addresses={self.addresses}"
        if self.socket is None and self.port is not None:
            self.command = f"{self.command} --port={self.port}"
        if self.socket is not None:
            self.command = f"{self.command} --socke={self.socket}"

    async def read_process(self):
        import pexpect
        self.process_callback.stop()
        try:
            # Simple way to catch everything is wait for a new line:
            #
            await self.process.expect("\n", async_=True)
        except pexpect.TIMEOUT:
            logger.debug("Reached timeout, getting back in the loop.")
        self.process_callback.start()

    async def launch(self):
        import pexpect
        import tornado.ioloop
        logger.info("Launching %s", self.command)
        parameters = {
            'command': self.command,
            'encoding': "utf-8"
        }
        if self.dir:
            parameters['cwd'] = self.dir
        if self.logfile is not None:
            parameters['logfile'] = self.logfile
        self.process = pexpect.spawn(**parameters)
        self.process.expect_list = expect_list_mp.__get__(self.process,
                                                          pexpect.spawn)
        await self.process.expect(
            [r"[Firenado server started successfully].*"], async_=True)
        self.process_callback = tornado.ioloop.PeriodicCallback(
            self.read_process,
            self.process_callback_time
        )
        self.process_callback.start()

    def send(self, line):
        logger.info("Sending line %s", line)
        self.process.sendline(line)

    def shutdown(self):
        logger.warning("Shutting down process launcher.")
        self.process.terminate(force=True)

    def is_alive(self):
        return self.process.isalive()


class TornadoLauncher(FirenadoLauncher):

    def __init__(self, **settings):
        super(TornadoLauncher, self).__init__(**settings)
        self.http_server = None
        self.application = None
        self.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN = firenado.conf.app[
            'wait_before_shutdown']
        if self.addresses is None or self.addresses == ['']:
            if firenado.conf.app['addresses']:
                self.addresses = firenado.conf.app['addresses']
            else:
                self.addresses = firenado.conf.app['default_addresses']
        if self.port is None:
            self.port = firenado.conf.app['port']

    def load(self):
        from .tornadoweb import TornadoApplication
        # TODO: Resolve module if doesn't exists
        if firenado.conf.app['pythonpath']:
            sys.path.append(firenado.conf.app['pythonpath'])
        self.application = TornadoApplication(**firenado.conf.app['settings'])

    def launch(self):
        import signal
        from tornado.ioloop import IOLoop
        import tornado.httpserver
        signal.signal(signal.SIGTERM, self.sig_handler)
        signal.signal(signal.SIGINT, self.sig_handler)
        if os.name == "posix":
            signal.signal(signal.SIGTSTP, self.sig_handler)
        self.http_server = tornado.httpserver.HTTPServer(self.application)
        if firenado.conf.app['xheaders'] is not None and type(
                firenado.conf.app['xheaders']) == bool:
            logger.debug("Setting http server xheaders as %s.",
                         firenado.conf.app['xheaders'])
            self.http_server.xheaders = firenado.conf.app['xheaders']
        if firenado.conf.app['xheaders'] is not None and type(
                firenado.conf.app['xheaders']) != bool:
            logger.warning("The xheaders defined in the application section"
                           "must be bool instead of %s. Ignoring the "
                           "configuration item.",
                           type(firenado.conf.app['xheaders']).__name__)
        listening_count = 0
        listening_what = "socket"
        if firenado.conf.app['socket'] or self.socket:
            from tornado.netutil import bind_unix_socket
            socket_path = firenado.conf.app['socket']
            if self.socket:
                socket_path = self.socket
            socket = bind_unix_socket(socket_path)
            self.http_server.add_socket(socket)
            logger.info("Firenado listening at socket %s",
                        socket.getsockname())
            listening_count += 1
        else:
            listening_what = "addresses:port"
            if len(self.addresses):
                for address in self.addresses:
                    if self.add_sockets(self.port, address):
                        listening_count += 1
            else:
                if self.add_sockets(self.port):
                    listening_count += 1
        if listening_count:
            if listening_count > 1:
                listening_what = f"{listening_what}s"
            logger.info("Firenado server started successfully. Listening at %s"
                        " %s.", listening_count, listening_what)
            if firenado.conf.app['process']['num_processes'] is not None:
                from tornado.process import fork_processes
                num_processes = firenado.conf.app['process']['num_processes']
                max_restarts = firenado.conf.app['process']['max_restarts']
                num_processes_alert = num_processes
                if num_processes == 0:
                    num_processes_alert = (
                        f"0 (assuming {tornado.process.cpu_count()} as cpu "
                        "count)")
                logger.info("Tornado set to start %s processes with %s max "
                            "restarts.", num_processes_alert, max_restarts)
                fork_processes(num_processes, max_restarts)
            IOLoop.current().start()
        else:
            logger.critical("Firenado unable to start.")
            sysexits.exit_fatal(sysexits.EX_SOFTWARE)

    def sig_handler(self, sig, _):
        """ Handle the signal sent to the process
        :param sig:  Signal set to the process
        :param _: Frame is not being used
        """
        from tornado.ioloop import IOLoop
        from tornado.process import task_id
        tid = task_id()
        pid = os.getpid()
        if tid is None:
            logger.warning("main process (pid %s) caught signal: %s", pid, sig)
        else:
            logger.warning("child %s (pid %s) caught signal: %s", tid, pid,
                           sig)
        IOLoop.current().add_callback_from_signal(self.shutdown)

    def add_sockets(self, port: int, address: str = None):
        from socket import gaierror
        from tornado.netutil import bind_sockets
        try:
            if address is None:
                address = "127.0.0.1"
            sockets = bind_sockets(port, address.strip())
            self.http_server.add_sockets(sockets)
            logger.info("Firenado listening at http://%s:%s.", address.strip(),
                        port)
            return True
        except gaierror:
            logger.error("Firenado failed to listen at http://%s:%s.",
                         address.strip(), port)
        except OSError:
            logger.error("Firenado failed to listen at http://%s:%s. Check if "
                         "another process is listening at this port or if you "
                         "provided a dns name and the correspondent ip in the "
                         "addresses configuration or if the machine owns the "
                         "ip Firenado will start to listen.", address, port)
        return False

    def shutdown(self):
        from tornado.ioloop import IOLoop
        from tornado.process import task_id
        import time

        def log_message(message: str, pid: int, tid: int = None):
            if tid is None:
                logger.info("main process (pid %s): %s", pid, message)
                return
            logger.info("child %s (pid %s): %s", tid, pid, message)

        tid = task_id()
        pid = os.getpid()
        log_message("stopping http server", pid, tid)
        for key, component in self.application.components.items():
            component.shutdown()
        self.http_server.stop()

        io_loop: IOLoop = IOLoop.current()

        if self.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN == 0:
            io_loop.stop()
            log_message("application is down", pid, tid)
            return

        log_message(f"shutdown in {self.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN}s"
                    " seconds", pid, tid)
        deadline = time.time() + self.MAX_WAIT_SECONDS_BEFORE_SHUTDOWN

        def stop_loop():
            now = time.time()
            if now < deadline:
                io_loop.add_timeout(now + 1, stop_loop)
                return
            io_loop.stop()
            log_message("application is down", pid, tid)
        stop_loop()



# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import tornadoweb
from cartola import xray
import functools
import inspect
from json.decoder import JSONDecodeError
import logging
from tornado import escape
from tornado.web import HTTPError
import warnings

logger = logging.getLogger(__name__)


try:
    import urlparse  # py2
except ImportError:
    import urllib.parse as urlparse  # py3


# TODO: I think we don't need a credential object. Looks like this is being
# changing to a generic metadata collection/holder to put on the session when
# the user is authenticated.
class Credential(object):

    def __init__(self):
        self.__authenticated = False

    def is_authenticated(self):
        return self.__authenticated

    def set_authenticated(self, authenticated):
        self.__authenticated = authenticated


# TODO: I think this needs to go
class Secured(object):
    @property
    def credential(self):
        """Returns the credential object """
        return self.__get_credential(self, '__credential_object', Credential)

    def __get_credential(self, obj, credential_attribute, credential_class):
        if not hasattr(obj, credential_attribute):
            setattr(obj, credential_attribute, credential_class())
        return getattr(obj, credential_attribute)


def secured(cls):
    for name, method in cls.__dict__.iteritems():
        if hasattr(method, "use_class"):
            # do something with the method and class
            print(name, cls)
    return cls


def identify(method):
    """ Decorator that gets all security data from session and adds to a
    credential.
    """
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        # TODO finish the session -> credential identification
        if self.session.get('UserId') is None:
            self.credential.set_authenticated(False)
        else:
            self.credential.set_authenticated(True)
        return method(self, *args, **kwargs)
    return wrapper


def only_xhr(method):
    """ Decorates a method in a handler to only accepts XMLHttpRequest
    requests.
    """
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if "X-Requested-With" in self.request.headers:
            if self.request.headers['X-Requested-With'] == "XMLHttpRequest":
                return method(self, *args, **kwargs)
        else:
            self.set_status(403)
            self.write("This is a XMLHttpRequest request only.")
    return wrapper


def default_get_current_user(self: tornadoweb.TornadoHandler):
    user_data = self.session.get("user")
    if user_data:
        # TODO: Let's think about this better, maybe we need some parameters
        # to figure out how to customize the conversion of user_data to
        # a python object. Meanwhile this is enough for most of the cases as
        # we're mostly echanging json around anyways.
        try:
            return escape.json_decode(user_data)
        except JSONDecodeError:
            logger.info("Cannot decode user_data %s to json. Rerturning raw "
                        "data.", user_data)
            return user_data
    return None


def default_class_authentication(self: tornadoweb.TornadoHandler):
    print("AAAA")
    print(self.get_current_user)
    print("AAAA")
    import firenado.conf
    if "login" in firenado.conf.app:
        warnings.warn("The \"login\" configuration in the application %s is"
                      "depreciated. Please replace the configuration app.login"
                      "to app.security.auth instead.", DeprecationWarning, 2)
    login_urls = self.get_rooted_path(
        firenado.conf.app['login']['urls']['default']
    )

    def must_login():
        decorators = xray.methods_decorators(self.__class__)
        method = self.request.method.lower()
        skip_auth_found = False
        skip_auth_variations = ["skip_auth", "security.skip_auth"]
        for decorator in filter(
                lambda d: "skip_auth" in d, decorators[method]):
            if decorator in skip_auth_variations:
                skip_auth_found = True
                break
        if skip_auth_found:
            logger.info("The handler at %s(method %s) is decoreated with "
                        "skip_auth. Bypassing authentication.",
                        self.request.path, method)
            return False
        login_url = self.get_rooted_path(
            firenado.conf.app['login']['urls']['default']
        )
        if login_url == self.request.path:
            logger.info("The path %s is a login url. Bypassing authentication"
                        ".", login_url)
            return False
        logger.info("Authentication is needed at %s.", self.request.path)
        return True

    if must_login():
        logger.info("Retrieving current user from the application.")
        user = self.current_user
        if not user:
            logger.info("The application has no current user logged.")
            login_url = self.get_rooted_path(
                firenado.conf.app['login']['urls']['default']
            )
            self.redirect(login_url)
            return
        logger.info("The application already has a current logged user: %s.",
                    user)


def new_autenticated(target=None, *args, **kwargs):
    """ This decorator will authenticate either a class or a method.
    If a class is decorated all http methods(i.e. get, post, delete, etc...)
    will be authenticated. There will be some options to skip the
    authentication process.
    If a method is decorated only that method will be authenticated.

    We could tell, using the authenticated decorator, that decorating a class
    is pessimistic, therefore every http requests will be blocked unless we
    skip the authentication process, and decorating a method is optimistic,
    because we'll block the method by individual basis.

    :param class|callable target:
    :param list args:
    :param dict kwargs:
    :return: A wrapped authenticaed class or method.
    """

    def build_class_authentication(cls: tornadoweb.TornadoHandler,
                                   *args, **kwargs):
        setattr(cls, "get_current_user", default_get_current_user)
        setattr(cls, "authenticate", default_class_authentication)
        print(cls)
        print(getattr(cls, "authenticate"))

        return cls

    if target is None:
        def f_wrapper(parametrized_target):
            if inspect.isclass(parametrized_target):
                return build_class_authentication(
                    parametrized_target, *args, **kwargs)
            else:
                @functools.wraps(parametrized_target)
                def parametrized_wrapper(self):
                    logger.info("buga")
                    print("buga")
                    print(self)
                    print("baaa")
                    if inspect.isclass(parametrized_target):
                        return build_class_authentication(
                            parametrized_target, *args, **kwargs)
                    print(parametrized_target)
                print("buuu")
                return parametrized_wrapper
        return f_wrapper


def authenticated(method):
    """ Decorator that checks if the user is authenticated.
    If not send the user to the login page."""
    def authenticate(self, *args, **kwargs):
        wrapped_method = kwargs.pop('wrapped_method')
        auth_url = kwargs.pop('url')
        is_authenticated = False
        # So if it has the current user than it is authenticated
        if self.current_user:
            is_authenticated = True
        if not is_authenticated:
            if self.request.method in ("GET", "HEAD"):
                self.redirect(auth_url)
                return
            raise HTTPError(403)
        return wrapped_method(self, *args, **kwargs)

    if hasattr(method, '__call__'):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            kwargs['wrapped_method'] = method
            try:
                url = self.get_login_url()
            except Exception:
                import firenado.conf
                # If url is not defined an exception will be raised
                url = firenado.conf.app['login']['urls']['default']
            # TODO: Why I'm not sending the full url anyways?
            if urlparse.urlsplit(url).scheme:
                # if login url is absolute, make next absolute too
                self.session.set('next_url', self.request.full_url())
            else:
                self.session.set('next_url', self.request.uri)
            kwargs['url'] = url
            return authenticate(self, *args, **kwargs)
        return wrapper
    else:
        def f_wrapper(par_method):
            import firenado.conf
            # Means someone added a parameter to the decorator
            url = firenado.conf.app['login']['urls'][method]

            @functools.wraps(par_method)
            def parametrized_wrapper(self, *args, **kwargs):
                kwargs['wrapped_method'] = par_method
                kwargs['url'] = url
                return authenticate(self, *args, **kwargs)
            return parametrized_wrapper
        return f_wrapper


def permissions(roles=None):
    if roles is None:
        roles = []

    def f_wrapper(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            return method(self, *args, **kwargs)
        return wrapper
    return f_wrapper



# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023 Flavio Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .sqlalchemy import with_session
import functools
import importlib
import logging


logger = logging.getLogger(__name__)


class FirenadoService:
    """ Base class to handle services. A Firenado service is usually connected
    to a handler or a service.
    The developer can add extra configuration using the configuration_service
    method and can get a data source from the data connected instance using
    either get_data_sources or get_data_source methods.
    """

    def __init__(self, consumer, data_source=None):
        self.consumer = consumer
        self.data_source = data_source
        self.configure_service()

    def configure_service(self):
        """ Use this method to add configurations to your service class.
        """
        pass

    def get_data_source(self, name):
        """ Returns a data source by its given name.
        """
        return self.get_data_sources()[name]

    def get_data_sources(self):
        """ If a data connected is returned then returns all data sources.
        If not returns None.

        :return: The data connected data sources
        """
        data_connected = self.data_connected
        if data_connected is not None:
            return data_connected.data_sources
        return None

    @property
    def data_connected(self):
        """ Will recurse over services until the data connected instance.
        If service has no consumer returns None.

        :return: The data connected object in the top of the hierarchy.
        """
        if self.consumer is None:
            return None
        from firenado.data import DataConnectedMixin
        if isinstance(self.consumer, DataConnectedMixin):
            return self.consumer
        invert_op = getattr(self.consumer, "get_data_connected", None)
        if callable(invert_op):
            return self.consumer.get_data_connected()
        return self.consumer.data_connected


def served_by(service, attribute_name=None):
    import warnings
    warnings.warn(
        "The \"firenado.service.served_by\" function is depreciated. "
        "Please use firenado.service.with_service instead.",
        DeprecationWarning, 2)
    return with_service(service, attribute_name)


def with_service(service, attribute_name=None):
    """ Decorator that connects a service to a service consumer.
    """

    def f_wrapper(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if isinstance(service, str):
                service_x = service.split('.')
                service_module = importlib.import_module(
                    '.'.join(service_x[:-1]))
                service_class = getattr(service_module, service_x[-1])
                service_name = service_x[-1]
            else:
                service_class = service
                service_name = service.__name__
            service_attribute = ''
            if attribute_name is None:
                first = True
                for s in service_name:
                    if s.isupper():
                        if first:
                            service_attribute = ''.join([
                                service_attribute, s.lower()])
                        else:
                            service_attribute = ''.join([
                                service_attribute, '_', s.lower()])
                    else:
                        service_attribute = ''.join([service_attribute, s])
                    first = False
            else:
                service_attribute = attribute_name
            must_set_service = False
            if not hasattr(self, service_attribute):
                must_set_service = True
            else:
                if getattr(self, service_attribute) is None:
                    must_set_service = True
            if must_set_service:
                setattr(self, service_attribute, service_class(self))
            return method(self, *args, **kwargs)

        return wrapper
    return f_wrapper


def sessionned(*args, **kwargs):
    import warnings
    warnings.warn(
        "The \"firenado.service.sessioned\" module is depreciated. "
        "Please use firenado.sqlalchemy.with_session  instead.",
        DeprecationWarning, 2)
    return with_session(*args, **kwargs)



# Copyright 2015-2023 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cartola import config
import errno
import firenado.conf
import functools
import logging
import sys

logger = logging.getLogger(__name__)


def configure(data_sources):
    """ Decorator that configures data sources on a data connected object.
    :param data_sources: List of data sources to be configured.
    """
    def f_wrapper(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            configure_data_sources(data_sources, self)
            return method(self, *args, **kwargs)
        return wrapper
    return f_wrapper


class DataConnectedMixin(object):
    """ Data connected objects has data sources. This mixin prepares an
    object to have data sources set to it and retrieved from it.

    Example:
    >>> class MyClass(..., DataConnectedMixin):
    """

    @property
    def data_sources(self):
        """ Returns all connectors registered to the data connected instance.
        """
        return get_data_sources(self, '__data_sources')

    def get_data_source(self, name):
        """ Returns a data source by its name.
        """
        return self.data_sources[name]

    def set_data_source(self, name, data_source):
        """ Add a data source to the data sources collection.
        """
        self.data_sources[name] = data_source


def get_data_sources(obj, data_sources_attribute):
    if not hasattr(obj, data_sources_attribute):
        setattr(obj, data_sources_attribute, {})
    return getattr(obj, data_sources_attribute)


class Connector(object):
    """ A connector will receive a data connection instance and provide
    a proper database connection to it.
    """

    def __init__(self, data_connected):
        self.__data_connected = data_connected

    def get_connection(self):
        """ Returns the configured and connected database connection.
        """
        return None

    def process_config(self, conf):
        """ Parse the configuration data provided by the firenado.conf engine.
        """
        return {}


class LdapConnector(Connector):

    def __init__(self, data_connected):
        super(LdapConnector, self).__init__(data_connected)


class RedisConnector(Connector):
    """ Connects a redis database to a data connected instance.
    """

    def __init__(self, data_connected):
        self.__connection = None
        super(RedisConnector, self).__init__(data_connected)
        self.__name = None

    def configure(self, name, conf):
        self.__name = name
        logger.info("Connecting to redis using the configuration: %s.", conf)
        import redis
        redis_conf = dict()
        redis_conf.update(conf)
        redis_conf.pop("connector")
        # TODO Handle connection error
        self.__connection = redis.Redis(**redis_conf)
        try:
            self.__connection.ping()
        except redis.ConnectionError as error:
            logger.fatal("Error trying to connect to redis: %s", error)
            sys.exit(errno.ECONNREFUSED)

    def get_connection(self):
        return self.__connection

    def process_config(self, conf):
        db_conf = {
            'connector': 'redis',
            'host': 'localhost',
            'port': 6379,
            'db': 0,
        }
        for key in conf:
            if key in ['db', 'host', 'port']:
                if key in ['db', 'port']:
                    db_conf[key] = int(conf[key])
                db_conf[key] = conf[key]
        return db_conf


class SqlalchemyConnector(Connector):
    from sqlalchemy.orm import Session
    """ Connects a sqlalchemy engine to a data connected instance.
    Sqlalchemy support a big variety of relational database backends. The
    connection returned by this handler contains an engine and session created
    by sqlalchemy and the database backend name.
    """

    def __init__(self, data_connected):
        super(SqlalchemyConnector, self).__init__(data_connected)
        self.__name = None
        self.__connection = {
            'backend': None,
            'session': {
                'autoflush': True,
                'expire_on_commit': True,
                'info': None
            }
        }
        self.__engine = None

    def configure(self, name, conf):
        self.__name = name
        from sqlalchemy import create_engine
        from sqlalchemy import exc, event, select
        # We will set the isolation level to READ UNCOMMITTED by default
        # to avoid the "cache" effect sqlalchemy has without this option.
        # Solution from: http://bit.ly/2bDq0Nv
        # TODO: Get the isolation level from data source conf
        engine_params = {
            'isolation_level': "READ UNCOMMITTED"
        }

        if "backend" in conf:
            if conf['backend'] == 'mysql':
                # Setting connection default connection timeout for mysql
                # backends as suggested on http://bit.ly/2bvOLxs
                engine_params['pool_recycle'] = 3600

        if "future" in conf:
            if conf['future']:
                engine_params['future'] = True

        if "pool" in conf:
            if "class" in conf['pool']:
                engine_params['pool_class'] = conf['pool']['class']
                if isinstance(engine_params['pool_class'], str):
                    engine_params['pool_class'] = config.get_from_string(
                        engine_params['pool_class'])
            if "isolation_level" in conf['pool']:
                engine_params['isolation_level'] = conf['pool'][
                    'isolation_level']
            if "max_overflow" in conf['pool']:
                engine_params['max_overflow'] = conf['pool']['max_overflow']
            if "pool_recycle" in conf['pool']:
                engine_params['pool_recycle'] = conf['pool']['pool_recycle']
            if "size" in conf['pool']:
                engine_params['pool_size'] = conf['pool']['size']

        if "session" in conf:
            if "autoflush" in conf['session']:
                self.__connection['session']['autoflush'] = conf['session'][
                    'autoflush']
            if "expire_on_commit" in conf['session']:
                self.__connection['session']['expire_on_commit'] = conf[
                    'session']['expire_on_commit']
            if "info" in conf['session']:
                self.__connection['session']['info'] = conf['session']['info']

        if "url" not in conf:
            logger.error("It is not possible to create sqlalchemy engine"
                         "for %s datasource. Configuration: %s.", self.__name,
                         conf)
        self.__engine = create_engine(conf['url'], **engine_params)

        @event.listens_for(self.__engine, "engine_connect")
        def ping_connection(connection, branch):
            # Adding ping connection event handler as described at the
            # pessimistic disconnect section of: http://bit.ly/2c8Sm2t
            logger.debug("Pinging sqlalchemy connection.")
            if branch:
                # "branch" refers to a sub-connection of a connection,
                # we don't want to bother pinging on these.
                logger.debug("The connection is a branch. There is no need to "
                             "ping those.")
                return
            # turn off "close with result".  This flag is only used with
            # "connectionless" execution, otherwise will be False in any case
            save_should_close_with_result = connection.should_close_with_result
            connection.should_close_with_result = False
            try:
                # run a SELECT 1.   use a core select() so that
                # the SELECT of a scalar value without a table is
                # appropriately formatted for the backend
                logger.debug("Testing sqlalchemy connection.")
                connection.scalar(select(1))
            except exc.DBAPIError as err:
                logger.warning(err)
                logger.warning("Firenado will try to reestablish the data "
                               "source connection.")
                # catch SQLAlchemy's DBAPIError, which is a wrapper
                # for the DBAPI's exception.  It includes a
                # .connection_invalidated attribute which specifies if this
                # connection is a "disconnect" condition, which is based on
                # inspection of the original exception by the dialect in use.
                if err.connection_invalidated:
                    # run the same SELECT again - the connection will
                    # re-validate itself and establish a new connection.
                    # The disconnect detection here also causes the whole
                    # connection pool to be invalidated so that all stale
                    # connections are discarded.
                    connection.scalar(select([1]))
                    logger.warning("Data source connection reestablished.")
                else:
                    raise
            finally:
                # restore "close with result"
                connection.should_close_with_result = (
                    save_should_close_with_result)
        logger.info("Connecting to the database using the engine: %s.",
                    self.__engine)
        self.__connection['backend'] = conf['backend']
        # will just happen during the handler execution

    def get_connection(self):
        return self.__connection

    def connect_engine(self):
        from sqlalchemy.exc import OperationalError
        try:
            self.__engine.connect()
        except OperationalError as op_error:
            logger.fatal("Error trying to connect to database: %s", op_error)
            sys.exit(errno.ECONNREFUSED)

    def get_a_session(self, **kwargs) -> Session:
        """ Return a session bind to the datasource engine.

        Default parameters based on: https://bit.ly/3MjWDzF
        :param dict kwargs:
        :key bool autoflush: Default to True
        :key bool expire_on_commit: Default to False
        :key dict info: Default to None
        :return Session:
        """
        autoflush = kwargs.get("autoflush", True)
        expire_on_commit = kwargs.get("expire_on_commit", True)
        info = kwargs.get("info")

        from sqlalchemy.orm import sessionmaker
        Session = sessionmaker(bind=self.__engine, autoflush=autoflush,
                               expire_on_commit=expire_on_commit, info=info)
        return Session()

    @property
    def backend(self):
        return self.__connection['backend']

    @property
    def engine(self):
        return self.__engine

    @property
    def session(self):
        return self.get_a_session()

    def process_config(self, conf):
        db_conf = {
            'type': 'sqlalchemy',
        }
        for key in conf:
            # TODO Handle other properties and create the url if needed.
            if key in ["db", "database", "dialect", "driver", "future", "host",
                       "pass", "password", "pool", "port", "session", "url",
                       "user", "username"]:
                index = key
                if index == "db":
                    index = "database"
                if index == "user":
                    index = "username"
                if index == "pass":
                    index = "password"
                db_conf[index] = conf[key]
        # TODO: Handler errors here
        if "url" in db_conf:
            db_conf['backend'] = db_conf['url'].split(':')[0].split('+')[0]
        else:
            url = ""
            if "dialect" in db_conf:
                db_conf['backend'] = db_conf['dialect']
                url = "%s" % db_conf.pop("dialect")
                if "driver" in db_conf:
                    url = "%s+%s" % (url, db_conf.pop("driver"))
            if "username" in db_conf:
                url = "%s://%s" % (url, db_conf.pop("username"))
                if "password" in db_conf:
                    from urllib.parse import quote
                    url = "%s:%s" % (url, quote(db_conf.pop("password")))
            if "host" in db_conf:
                url = "%s@%s" % (url, db_conf.pop("host"))
                if "port" in db_conf:
                    url = "%s:%s" % (url, db_conf.pop("port"))
            if "database" in db_conf:
                url = "%s/%s" % (url, db_conf.pop("database"))
            db_conf['url'] = url
        return db_conf


def config_to_data_source(name, conf, data_connected):
    """ Convert a data source conf to its respective data source. We need
    a data connected to use while instantiating the data source.
    :param name: Datasource name
    :param conf: A data source confuration item
    :param data_connected: A data connected object
    :return: Connector
    """
    connector_conf = firenado.conf.data['connectors'][conf['connector']]
    # TODO: Test if handler was returned None. An error occurred.
    handler_class = config.get_from_module(connector_conf['module'],
                                           connector_conf['class'])
    data_source_instance = handler_class(data_connected)
    conf = data_source_instance.process_config(conf)
    data_source_instance.configure(name, conf)
    return data_source_instance


def configure_data_sources(data_sources, data_connected):
    """ Configure all data sources from configuration and set to the data
    connected.
    :param data_sources: List of data sources to be configured
    :param data_connected: Data connected object where the data sources will
    be configured.
    """
    if isinstance(data_sources, str):
        if data_sources in firenado.conf.data['sources']:
            logger.debug("Found data source [%s] in the list. Preceding with "
                         "the configuration process.", data_sources)
            conf = firenado.conf.data['sources'][data_sources]
            data_source_instance = config_to_data_source(
                data_sources, conf, data_connected)
            data_connected.set_data_source(data_sources, data_source_instance)
        else:
            logger.fatal("It was not possible to find [%s] in the list of "
                         "available data sources. Please fix the firenado "
                         "configuration file. Sometimes that could be only a "
                         "typo in one of app data sources to be created. Look "
                         "at app.data.sources list.", data_sources)
            sys.exit(errno.ENOKEY)
        # TODO: Testing the connection
        # Without that the error will just happen during the handler execution
    elif isinstance(data_sources, list):
        for data_source in data_sources:
            configure_data_sources(data_source, data_connected)
    # TODO Throw an error here if it is not string or list
